@@ -26,6 +26,7 @@ import time
 from typing import TYPE_CHECKING, Any, TypeVar, cast
 from uuid import UUID
 
+from bleak.backends.characteristic import BleakGATTCharacteristic
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 from bleak.exc import BleakError
@@ -49,7 +50,7 @@ from aiohomekit.model import (
     Transport,
 )
 from aiohomekit.model.characteristics import Characteristic, CharacteristicPermissions
-from aiohomekit.model.services import ServicesTypes
+from aiohomekit.model.services import Service, ServicesTypes
 from aiohomekit.pdu import OpCode, PDUStatus, decode_pdu, encode_pdu
 from aiohomekit.protocol import get_session_keys
 from aiohomekit.protocol.statuscodes import HapStatusCode
@@ -75,6 +76,7 @@ from .structs import (
     Characteristic as CharacteristicTLV,
     ProtocolParams,
     ProtocolParamsTLV,
+    Service as ServiceTLV,
 )
 from .values import from_bytes, to_bytes
 
@@ -95,6 +97,8 @@ NEVER_TIME = -AVAILABILITY_INTERVAL
 SERVICE_INSTANCE_ID = "E604E95D-A759-4817-87D3-AA005083A0D1"
 SERVICE_INSTANCE_ID_UUID = UUID(SERVICE_INSTANCE_ID)
 
+SERVICE_SIGNATURE_UUID = UUID(CharacteristicsTypes.SERVICE_SIGNATURE)
+
 SKIP_SYNC_SERVICES = {
     ServicesTypes.PAIRING,
     ServicesTypes.TRANSFER_TRANSPORT_MANAGEMENT,
@@ -159,6 +163,12 @@ def restore_connection_and_resume(func: WrapFuncType) -> WrapFuncType:
         try:
             return await func(self, *args, **kwargs)
         finally:
+            logger.debug(
+                "%s: Finished %s, checking for subscription restore: %s",
+                self.name,
+                func.__name__,
+                self._restore_pending,
+            )
             if not self._shutdown and self._restore_pending:
                 await self._async_restore_subscriptions()
 
@@ -354,17 +364,19 @@ class BlePairing(AbstractPairing):
         self._broadcast_notifications = set()
         self._restore_pending = False
 
-    async def _ensure_connected(self, attempts: int | None = None):
-        """Ensure that we are connected to the accessory."""
+    async def _ensure_connected(self, attempts: int | None = None) -> bool | None:
+        """Ensure that we are connected to the accessory.
+
+        Returns True if we had to make the connection,
+        returns False if we were already connected or shutdown.
+        """
         assert self._config_lock.locked(), "_config_lock Should be locked"
-        if self.client and self.client.is_connected:
-            return
+        if self._shutdown or (self.client and self.client.is_connected):
+            return False
         async with self._connection_lock:
-            if self._shutdown:
-                return
             # Check again while holding the lock
-            if self.client and self.client.is_connected:
-                return
+            if self._shutdown or (self.client and self.client.is_connected):
+                return False
             if not self.device and (
                 discovery := await self.controller.async_get_discovery(
                     self.address, DISCOVER_TIMEOUT
@@ -385,12 +397,16 @@ class BlePairing(AbstractPairing):
                 ble_device_callback=lambda: self.device,
                 max_attempts=attempts,
             )
+            return True
 
     async def _async_start_notify(self, iid: int) -> None:
-        char = self.accessories.aid(1).characteristics.iid(iid)
+        char = self.accessories.aid(BLE_AID).characteristics.iid(iid)
 
         # Find the GATT Characteristic object for this iid
-        endpoint = self.client.get_characteristic(char.service.type, char.type)
+        if char.handle:
+            endpoint = self.client.get_characteristic_by_handle(char.handle)
+        else:
+            endpoint = self.client.get_characteristic(char.service.type, char.type)
 
         # We only want to allow one in flight read
         # and one pending read at a time since there
@@ -407,11 +423,15 @@ class BlePairing(AbstractPairing):
                     # Client disconnected
                     return
                 logger.debug("%s: Retrieving event for iid: %s", self.name, iid)
-                if results := await self._get_characteristics_without_retry(
-                    [(BLE_AID, iid)]
-                ):
-                    for listener in self.listeners:
-                        listener(results)
+                await self._get_characteristics_without_retry(
+                    [(BLE_AID, iid)], notify_listeners=True
+                )
+                # After a char has changed we need to check if the
+                # GSN has changed as well so we don't reconnect
+                # to the accessory if we don't need to
+                protocol_param = await self._get_all_protocol_params()
+                if protocol_param:
+                    self.description.state_num = protocol_param.state_number
 
         def _callback(id: int, data: bytes) -> None:
             logger.debug("%s: Received event for iid=%s: %s", self.name, iid, data)
@@ -467,8 +487,7 @@ class BlePairing(AbstractPairing):
                 self.rssi,
             )
             try:
-                protocol_param = await self.get_all_protocol_params()
-                results = await self.get_characteristics(list(self.subscriptions))
+                protocol_param = await self._process_disconnected_events_with_retry()
             except (
                 AccessoryDisconnectedError,
                 *BLEAK_EXCEPTIONS,
@@ -482,12 +501,23 @@ class BlePairing(AbstractPairing):
                 )
                 return
 
-            for listener in self.listeners:
-                listener(results)
-
             if protocol_param:
                 self.description.state_num = protocol_param.state_number
 
+    @operation_lock
+    @retry_bluetooth_connection_error()
+    @restore_connection_and_resume
+    async def _process_disconnected_events_with_retry(
+        self,
+    ) -> ProtocolParams | None:
+        accessory_chars = self.accessories.aid(BLE_AID).characteristics
+        protocol_param = await self._get_all_protocol_params()
+        await self._get_characteristics_while_connected(
+            [accessory_chars.iid(iid) for _, iid in self.subscriptions],
+            notify_listeners=True,
+        )
+        return protocol_param
+
     def _async_notification(self, data: HomeKitEncryptedNotification) -> None:
         """Receive a notification from the accessory."""
         if not self._broadcast_decryption_key:
@@ -544,7 +574,7 @@ class BlePairing(AbstractPairing):
             )
             # We had a successful decrypt, so we can update the state_num
             self.description.state_num = gsn
-            char = self.accessories.aid(1).characteristics.iid(iid)
+            char = self.accessories.aid(BLE_AID).characteristics.iid(iid)
 
             results = {(BLE_AID, iid): {"value": from_bytes(char, value)}}
             logger.debug("%s: Received notification: results = %s", self.name, results)
@@ -559,7 +589,7 @@ class BlePairing(AbstractPairing):
 
     def _async_get_service_signature_char(self) -> Characteristic | None:
         """Get the service signature characteristic."""
-        info = self.accessories.aid(1).services.first(
+        info = self.accessories.aid(BLE_AID).services.first(
             service_type=ServicesTypes.PROTOCOL_INFORMATION
         )
         if not info:
@@ -576,34 +606,70 @@ class BlePairing(AbstractPairing):
 
     async def _async_set_broadcast_encryption_key(self) -> None:
         """Get the broadcast key for the accessory."""
-        hap_char = self._async_get_service_signature_char()
-        if not hap_char:
-            return
-        service_iid = hap_char.service.iid
-        logger.debug(
-            "%s: Setting broadcast key for service_iid: %s",
-            self.name,
-            service_iid,
-        )
-        try:
-            await self._async_request_under_lock(
-                OpCode.PROTOCOL_CONFIG,
-                hap_char,
-                GENERATE_BROADCAST_KEY_PAYLOAD,
-                iid=service_iid,
-            )
-        except PDUStatusError:
-            logger.exception(
-                "%s: Failed to set broadcast key, try un-paring and re-pairing the accessory.",
+        logger.debug("%s: Setting broadcast encryption key", self.name)
+        if self._ble_request_lock.locked():
+            logger.debug(
+                "%s: Waiting ble request lock to set broadcast encryption key",
                 self.name,
             )
-            return
+        async with self._ble_request_lock:
+            hap_char = self._async_get_service_signature_char()
+            if not hap_char:
+                return
+            service_iid = hap_char.service.iid
+            logger.debug(
+                "%s: Setting broadcast key for service_iid: %s",
+                self.name,
+                service_iid,
+            )
+            try:
+                await self._async_request_under_lock(
+                    OpCode.PROTOCOL_CONFIG,
+                    hap_char,
+                    GENERATE_BROADCAST_KEY_PAYLOAD,
+                    iid=service_iid,
+                )
+            except PDUStatusError:
+                logger.exception(
+                    "%s: Failed to set broadcast key, try un-paring and re-pairing the accessory.",
+                    self.name,
+                )
+        long_term_pub_key_hex: str = self.pairing_data["iOSDeviceLTPK"]
+        long_term_pub_key_bytes = bytes.fromhex(long_term_pub_key_hex)
+        self._broadcast_decryption_key = BroadcastDecryptionKey(
+            self._derive(long_term_pub_key_bytes, b"Broadcast-Encryption-Key")
+        )
+
+    async def _read_signature(
+        self,
+        char: BleakGATTCharacteristic,
+        op_code: OpCode,
+        iid: int,
+        tlv_struct: ServiceTLV | CharacteristicTLV,
+    ) -> dict[str, Any]:
+        """Read the signature for the given characteristic."""
+        tid = random.randint(1, 254)
+        for data in encode_pdu(op_code, tid, iid):
+            await self.client.write_gatt_char(
+                char,
+                data,
+                "write-without-response" not in char.properties,
+            )
+
+        payload = await self.client.read_gatt_char(char)
+
+        status, _, signature = decode_pdu(tid, payload)
+        if status != PDUStatus.SUCCESS:
+            return {}
+
+        return tlv_struct.decode(signature).to_dict()
 
     async def _async_fetch_gatt_database(self) -> Accessories:
         logger.debug("%s: Fetching GATT database; rssi=%s", self.name, self.rssi)
         accessory = Accessory()
         accessory.aid = 1
         # Never use the cache when fetching the GATT database
+        services_to_link: list[tuple[Service, list[int]]] = []
         services = await self.client.get_services()
         for service in services:
             ble_service_char = service.get_characteristic(SERVICE_INSTANCE_ID_UUID)
@@ -629,8 +695,23 @@ class BlePairing(AbstractPairing):
             s = accessory.add_service(normalize_uuid(service.uuid))
             s.iid = service_iid
 
+            service_signature_char = service.get_characteristic(SERVICE_SIGNATURE_UUID)
+            if service_signature_char:
+                decoded_service = await self._read_signature(
+                    service_signature_char,
+                    OpCode.SERV_SIG_READ,
+                    service_iid,
+                    ServiceTLV,
+                )
+                if "linked" in decoded_service:
+                    services_to_link.append((s, decoded_service["linked"]))
+                logger.debug(
+                    "%s: service: %s decoded: %s", self.name, service, decoded_service
+                )
+
             for char in service.characteristics:
-                if normalize_uuid(char.uuid) == SERVICE_INSTANCE_ID:
+                normalized_uuid = normalize_uuid(char.uuid)
+                if normalized_uuid == SERVICE_INSTANCE_ID:
                     continue
 
                 iid = await self.client.get_characteristic_iid(char)
@@ -638,27 +719,9 @@ class BlePairing(AbstractPairing):
                     logger.debug("%s: No iid for %s", self.name, char.uuid)
                     continue
 
-                tid = random.randint(1, 254)
-                for data in encode_pdu(
-                    OpCode.CHAR_SIG_READ,
-                    tid,
-                    iid,
-                ):
-                    await self.client.write_gatt_char(
-                        char,
-                        data,
-                        "write-without-response" not in char.properties,
-                    )
-
-                payload = await self.client.read_gatt_char(char)
-
-                status, _, signature = decode_pdu(tid, payload)
-                if status != PDUStatus.SUCCESS:
-                    continue
-
-                decoded = CharacteristicTLV.decode(signature).to_dict()
-                normalized_uuid = normalize_uuid(char.uuid)
-
+                decoded = await self._read_signature(
+                    char, OpCode.CHAR_SIG_READ, iid, CharacteristicTLV
+                )
                 if normalized_uuid == CharacteristicsTypes.IDENTIFY:
                     # Workaround for older eve v1 devices which has a broken identify characteristic
                     # that presents identify as data.
@@ -685,6 +748,12 @@ class BlePairing(AbstractPairing):
                 if "broadcast_events" in decoded:
                     hap_char.broadcast_events = decoded["broadcast_events"]
 
+        # Link services after we are done since we need to have all services
+        # in the accessory to link them.
+        for service, linked in services_to_link:
+            for iid in linked:
+                service.add_linked_service(accessory.services.iid(iid))
+
         accessories = Accessories()
         accessories.add_accessory(accessory)
         logger.debug("%s: Completed fetching GATT database", self.name)
@@ -726,7 +795,7 @@ class BlePairing(AbstractPairing):
     async def _populate_char_values(self, config_changed: bool) -> None:
         """Populate the values of all characteristics."""
         chars: list[Characteristic] = []
-        for service in self.accessories.aid(1).services:
+        for service in self.accessories.aid(BLE_AID).services:
             if service.type in SKIP_SYNC_SERVICES:
                 continue
             if (
@@ -758,13 +827,6 @@ class BlePairing(AbstractPairing):
         if protocol_params:
             self.description.state_num = protocol_params.state_number
 
-    @operation_lock
-    @retry_bluetooth_connection_error()
-    @restore_connection_and_resume
-    async def get_all_protocol_params(self) -> ProtocolParams | None:
-        """Get the global state number."""
-        return await self._get_all_protocol_params()
-
     async def _get_all_protocol_params(self) -> ProtocolParams | None:
         """Get the current protocol params number."""
         hap_char = self._async_get_service_signature_char()
@@ -837,7 +899,7 @@ class BlePairing(AbstractPairing):
         Older code did not save the handle, so if we have no handles
         we need to re-fetch the gatt database.
         """
-        for service in self.accessories.aid(1).services:
+        for service in self.accessories.aid(BLE_AID).services:
             for char in service.characteristics:
                 if char.handle is not None:
                     return False
@@ -851,17 +913,17 @@ class BlePairing(AbstractPairing):
             if self._shutdown:
                 return
 
-            was_connected = self.client and self.client.is_connected
-            self._restore_pending |= not was_connected
             self._tried_to_connect_once = True
+
+            made_connection = await self._ensure_connected(attempts)
+
             logger.debug(
-                "%s: Populating accessories and characteristics: was_connected=%s restore_pending=%s",
+                "%s: Populating accessories and characteristics: made_connection=%s restore_pending=%s",
                 self.name,
-                was_connected,
+                made_connection,
                 self._restore_pending,
             )
-
-            await self._ensure_connected(attempts)
+            self._restore_pending |= made_connection
 
             if was_locked and not force_update:
                 # No need to do it twice if we already have the data
@@ -903,7 +965,7 @@ class BlePairing(AbstractPairing):
         self, subscriptions: list[tuple[int, int]]
     ) -> None:
         """Subscribe to broadcast events."""
-        accessory_chars = self.accessories.aid(1).characteristics
+        accessory_chars = self.accessories.aid(BLE_AID).characteristics
         to_subscribe: list[Characteristic] = []
         for _, iid in subscriptions:
             hap_char = accessory_chars.iid(iid)
@@ -944,19 +1006,12 @@ class BlePairing(AbstractPairing):
         if not self._restore_pending or not self.client or not self.client.is_connected:
             return
 
-        logger.debug("%s: Setting broadcast encryption key", self.name)
-        async with self._ble_request_lock:
-            await self._async_set_broadcast_encryption_key()
-        long_term_pub_key_hex: str = self.pairing_data["iOSDeviceLTPK"]
-        long_term_pub_key_bytes = bytes.fromhex(long_term_pub_key_hex)
-        self._broadcast_decryption_key = BroadcastDecryptionKey(
-            self._derive(long_term_pub_key_bytes, b"Broadcast-Encryption-Key")
-        )
-
         if not self.subscriptions:
+            logger.debug("%s: No subscriptions to restore", self.name)
             self._restore_pending = False
             return
 
+        await self._async_set_broadcast_encryption_key()
         subscriptions = list(self.subscriptions)
         logger.debug(
             "%s: Connected, resuming subscriptions: %s; rssi=%s",
@@ -967,6 +1022,11 @@ class BlePairing(AbstractPairing):
         await self._async_subscribe_broadcast_events(subscriptions)
         await self._async_start_notify_subscriptions(subscriptions)
         self._restore_pending = False
+        # After we have restored subscriptions, we need to read
+        # the state number again to make sure we are in sync
+        protocol_param = await self._get_all_protocol_params()
+        if protocol_param:
+            self.description.state_num = protocol_param.state_number
 
     async def _async_start_notify_subscriptions(
         self, subscriptions: list[tuple[int, int]]
@@ -1018,7 +1078,7 @@ class BlePairing(AbstractPairing):
             ]
         )
 
-        info = self.accessories.aid(1).services.first(
+        info = self.accessories.aid(BLE_AID).services.first(
             service_type=ServicesTypes.PAIRING
         )
         char = info[CharacteristicsTypes.PAIRING_PAIRINGS]
@@ -1058,15 +1118,17 @@ class BlePairing(AbstractPairing):
     async def _get_characteristics_without_retry(
         self,
         characteristics: list[tuple[int, int]],
+        notify_listeners: bool = False,
     ) -> dict[tuple[int, int], dict[str, Any]]:
-        accessory_chars = self.accessories.aid(1).characteristics
+        accessory_chars = self.accessories.aid(BLE_AID).characteristics
         return await self._get_characteristics_while_connected(
-            [accessory_chars.iid(iid) for _, iid in characteristics]
+            [accessory_chars.iid(iid) for _, iid in characteristics], notify_listeners
         )
 
     async def _get_characteristics_while_connected(
         self,
         characteristics: list[Characteristic],
+        notify_listeners: bool = False,
     ) -> dict[tuple[int, int], dict[str, Any]]:
         if logger.isEnabledFor(logging.DEBUG):
             logger.debug(
@@ -1115,7 +1177,7 @@ class BlePairing(AbstractPairing):
                 )
 
                 try:
-                    results[(BLE_AID, char.iid)] = {"value": from_bytes(char, decoded)}
+                    value = from_bytes(char, decoded)
                 except struct.error as ex:
                     logger.debug(
                         "%s: Failed to decode characteristic for %s from %s: %s",
@@ -1124,6 +1186,19 @@ class BlePairing(AbstractPairing):
                         decoded,
                         ex,
                     )
+                    continue
+
+                result_key = (BLE_AID, char.iid)
+                result_value = {"value": value}
+                results[result_key] = result_value
+
+                if notify_listeners:
+                    # Since it can take a while to read all the characteristics
+                    # we want to notify the listeners as soon as we have the
+                    # value for each characteristic.
+                    single_results = {result_key: result_value}
+                    for listener in self.listeners:
+                        listener(single_results)
 
         return results
 
@@ -1140,7 +1215,7 @@ class BlePairing(AbstractPairing):
             characteristics,
             self.rssi,
         )
-        accessory_chars = self.accessories.aid(1).characteristics
+        accessory_chars = self.accessories.aid(BLE_AID).characteristics
         async with self._ble_request_lock:
 
             for aid, iid, value in characteristics:
@@ -1193,6 +1268,8 @@ class BlePairing(AbstractPairing):
             return
         logger.debug("%s: subscribing to %s", self.name, new_chars)
         await self._populate_accessories_and_characteristics()
+        if not self._broadcast_decryption_key:
+            await self._async_set_broadcast_encryption_key()
         await self._async_subscribe_broadcast_events(new_chars)
         await self._async_start_notify_subscriptions(new_chars)
 
@@ -1200,7 +1277,7 @@ class BlePairing(AbstractPairing):
         pass
 
     async def identify(self):
-        info = self.accessories.aid(1).services.first(
+        info = self.accessories.aid(BLE_AID).services.first(
             service_type=ServicesTypes.ACCESSORY_INFORMATION
         )
         char = info[CharacteristicsTypes.IDENTIFY]
@@ -1244,7 +1321,7 @@ class BlePairing(AbstractPairing):
             ]
         )
 
-        info = self.accessories.aid(1).services.first(
+        info = self.accessories.aid(BLE_AID).services.first(
             service_type=ServicesTypes.PAIRING
         )
         char = info[CharacteristicsTypes.PAIRING_PAIRINGS]
@@ -1286,7 +1363,7 @@ class BlePairing(AbstractPairing):
             ]
         )
 
-        info = self.accessories.aid(1).services.first(
+        info = self.accessories.aid(BLE_AID).services.first(
             service_type=ServicesTypes.PAIRING
         )
         char = info[CharacteristicsTypes.PAIRING_PAIRINGS]
